@@ -12,13 +12,15 @@ import pandas
 import numpy
 import time
 import typing
+import tempfile
 from collections import namedtuple
 import yaml
 
 from .proto_handlers import load_protocol_handler
 from .utils.connect import connect_to_database
 from .utils.dbconn import sql_enquote_id
-from .utils.file_utils import write_stream_with_read_on_close, connect_to_filesystem, find_url_extension
+from .utils.file_utils import write_stream_with_read_on_close, connect_to_filesystem
+from .utils.format_sniffer import find_url_extension, detect_format, looks_like_text_file
 from .utils.nosql import connect_to_nosql
 from .utils.packages import install_packages
 
@@ -248,7 +250,7 @@ class DSLibrary(object):
         """
         return self.open_resource(f"model-binary/{part}" if part else "model-binary", mode=mode)
 
-    def open_run_data(self, filename: str, mode: str='rb') -> io.RawIOBase:
+    def open_run_data(self, filename: str, mode: str='rb', **kwargs) -> io.RawIOBase:
         """
         Open a stream to or from file-like storage which is local to the current pipeline execution context.
 
@@ -283,10 +285,10 @@ class DSLibrary(object):
             return connect_to_filesystem(uri=uri, for_write=for_write, **kwargs)
         if system_type == "sql":
             library = kwargs.pop("library", None)
-            return connect_to_database(uri, library=library, for_write=for_write, **kwargs)
+            return connect_to_database(uri=uri, library=library, for_write=for_write, **kwargs)
         if system_type == "nosql":
             library = kwargs.pop("library", None)
-            return connect_to_nosql(uri, library=library, for_write=for_write, **kwargs)
+            return connect_to_nosql(uri=uri, library=library, for_write=for_write, **kwargs)
         raise ValueError(f"Unsupported system_type: {system_type}")
 
     def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
@@ -354,7 +356,7 @@ class DSLibrary(object):
         """
         Save a piece of trackable information about this run.
         """
-        if self._mlflow_metrics:
+        if self._mlflow_metrics or self._mlflow_all:
             if mlflow:
                 mlflow.log_metric(metric_name, metric_value, step)
             return
@@ -483,7 +485,8 @@ class DSLibrary(object):
         use_mlflow = self._mlflow_all
         if use_mlflow:
             mlflow.start_run()
-            self._run_id = mlflow.active_run().info.run_id
+            a_r = mlflow.active_run()
+            self._run_id = a_r.info.run_id if a_r else None
         else:
             # generate a run ID
             self._run_id = str(uuid.uuid4())
@@ -579,6 +582,7 @@ class DSLibrary(object):
                         for f in ["sep"]:
                             if f in format_options:
                                 chk_args[f] = format_options[f]
+                        # FIXME use load_dataframe() here to get sniffing/etc.
                         df0 = PANDAS_READ_FORMATS[df_format](f_r, **chk_args)
                         # verify column names match
                         cols_l = set(df0.columns)
@@ -794,20 +798,64 @@ class DSLibrary(object):
             return pandas.DataFrame(rows)
         # TODO we should offer a validation option that checks the data types against the schema
         # format can be specified, or file extension will be used
-        file_extension = find_url_extension(uri).strip(".")
+        file_extension = find_url_extension(uri)
         if not file_extension:
             self._default_format_for_aliases(resource_name, open_args)
-        # options for read function can be given
-        format = open_args.pop("format", None) or file_extension
         format_options = open_args.pop("format_options", None) or {}
-        # various formats and their read function
-        if format in PANDAS_READ_FORMATS:
-            if open_args.get("run_data"):
-                # read from run data
-                with self.open_run_data(resource_name, mode='rb') as f_r:
-                    return PANDAS_READ_FORMATS[format](f_r, **format_options)
+        format = open_args.pop("format", None)
+        if format:
+            format_options = dict(format_options)
+            format_options["format"] = format
+        # sample-taker
+        use_opener = self.open_run_data if open_args.pop("run_data", None) else self.open_resource
+        def sampler():
+            sample_size = open_args.pop("sample_size", None) or 30000
+            if sample_size < 0:
+                return
+            with use_opener(uri, mode='rb', **open_args) as f_r:
+                return f_r.read(sample_size)
+        '''
+        # TODO we could measure the size, but this doesn't add much to our format sniffing and could cause problems on
+        #   some data sources
+        def sizer():
             with self.open_resource(resource_name, mode='rb', **kwargs) as f_r:
-                return PANDAS_READ_FORMATS[format](f_r, **format_options)
+                if not hasattr(f_r, "seek") or not f_r.seekable():
+                    return
+                f_r.seek(0, 2)
+                return f_r.tell()
+        '''
+        # sniff format
+        fallback_to_text = open_args.pop("fallback_to_text", False)
+        format_info = detect_format(
+            uri, sampler=sampler, sizer=None, supplied_options=format_options, fallback_to_text=fallback_to_text
+        )
+        format = format_info.format
+        read_args = format_info.read_args
+        # various formats and their read function
+        use_dask = open_args.pop("dask", None)
+        if use_dask:
+            read_method = DASK_READ_FORMATS.get(format, PANDAS_READ_FORMATS.get(format))
+            if format == "csv" and format in DASK_READ_FORMATS:
+                # dask reads 'sample' bytes to determine dtypes, default is 256k, which is often too small
+                #  - 10M is also bound to be too small of course, but it's somewhat less likely
+                if "sample" not in read_args:
+                    read_args["sample"] = 10000000
+                # without this, data with ints at the front and floats at the back raises an exception
+                #  - this solves one typical problem from the limited dtype sampling above
+                if "assume_missing" not in read_args:
+                    read_args["assume_missing"] = True
+        else:
+            read_method = PANDAS_READ_FORMATS.get(format)
+        if read_method:
+            with use_opener(uri, mode='rb', **open_args) as f_r:
+                if format == "xlsx" and not f_r.seekable():
+                    # the XLSX reader requires its stream to be seekable so we download to a temporary file
+                    with tempfile.NamedTemporaryFile(mode='wb') as f_tmp:
+                        f_tmp.write(f_r.read())
+                        f_tmp.flush()
+                        with open(f_tmp.name, mode='rb') as f_r2:
+                            return read_method(f_r2, **read_args)
+                return read_method(f_r, **read_args)
         # TODO support read from folder, i.e. folder of CSVs, parquet, etc.
         raise DSLibraryException(f"file type not supported: {format}")
 
@@ -931,8 +979,10 @@ LOCAL_SPECS_SCHEMA = {
 
 PANDAS_READ_FORMATS = {
     "csv": pandas.read_csv,
-    "tab": pandas.read_csv,
+    # TODO should we use json_normalize() or read_json() here, or add a switch?
+    #  - note that json and yaml are currently different
     "json": pandas.read_json,
+    "yaml": lambda fh, **kwargs: pandas.json_normalize(yaml.safe_load(fh), **kwargs),
     "parquet": pandas.read_parquet,
     "xlsx": pandas.read_excel,
     "xls": pandas.read_excel,
@@ -950,3 +1000,15 @@ PANDAS_WRITE_FORMATS = {
     "hdf": "to_hdf",
     "fwf": "to_fwf"
 }
+
+try:
+    import dask.dataframe
+    DASK_READ_FORMATS = {
+        "csv": dask.dataframe.read_csv,
+        "json": dask.dataframe.read_json,
+        "parquet": dask.dataframe.read_parquet,
+        "hdf": dask.dataframe.read_hdf,
+        "fwf": dask.dataframe.read_fwf
+    }
+except ImportError:
+    DASK_READ_FORMATS = {}
