@@ -6,9 +6,10 @@ import numpy
 import pandas
 import shutil
 import tempfile
+import yaml
 
 from dslibrary import DSLibrary
-from dslibrary.front import EVALUATION_RESULT_ALIAS, METRICS_ALIAS
+from dslibrary.front import EVALUATION_RESULT_ALIAS, METRICS_ALIAS, DSLibraryException
 from dslibrary.metadata import Metadata
 from dslibrary.transport.to_local import DSLibraryLocal
 
@@ -206,7 +207,8 @@ class TestFront(unittest.TestCase):
                 return io.BytesIO(b"x\n1\n2")
         df = MyInst().load_dataframe("x", uri="default.csv")
         df = MyInst(spec={"inputs": {"x": {"uri": "override.csv"}}}).load_dataframe("x", uri="default.csv")
-        self.assertEqual(log, ['default.csv', 'override.csv'])
+        self.assertEqual(log[0], 'default.csv')
+        self.assertEqual(log[-1], 'override.csv')
 
     def test_load_dataframe__from_sql(self):
         log = []
@@ -313,3 +315,165 @@ class TestFront(unittest.TestCase):
         df2 = dsl.load_dataframe("x.csv", run_data=True)
         self.assertEqual(list(df2.x), list(df.x))
         self.assertEqual(dsl.read_run_data("x.csv"), b'x\n1\n2\n')
+
+    def test_open_args_passed_through_to_opener(self):
+        log = []
+        inst = self
+        class MyInst(DSLibrary):
+            def _opener(self, path: str, mode: str, **kwargs):
+                log.append(path)
+                inst.assertEqual(kwargs, {'arg1': 123})
+                return io.BytesIO(b"abc")
+        # just a filename
+        fh = MyInst().open_resource("x", arg1=123, format="xyz")
+        assert fh.read() == b"abc"
+        # uri overrides name
+        fh = MyInst().open_resource("x", uri="default.csv", arg1=123, format="xyz")
+        assert fh.read() == b"abc"
+        # input mapping maps 'x' to something else
+        fh = MyInst(spec={"inputs": {"x": {"uri": "override.csv"}}}).open_resource("x", uri="default.csv", arg1=123, format="xyz")
+        assert fh.read() == b"abc"
+        # verify filenames
+        self.assertEqual(log, ['x', 'default.csv', 'override.csv'])
+
+    def test_setup_code_paths(self):
+        """
+        You can cause 'sys.path' to be extended to include particular folders.
+        This is normally done using an environment variable.  Here we are just testing the method that applies the
+        changes.
+        """
+        inst = DSLibrary()
+        inst._spec = {"code_paths": ["a", "b"]}
+        mock_paths = []
+        with mock.patch("sys.path", mock_paths):
+            inst._setup_code_paths()
+        self.assertEqual(mock_paths, ["a", "b"])
+
+    def test_abstract_base_methods(self):
+        """
+        The methods that have to be filled in raise exceptions when called.
+        """
+        inst = DSLibrary()
+        self.assertRaises(DSLibraryException, lambda: inst._opener("path", mode='r'))
+        self.assertRaises(DSLibraryException, lambda: inst.open_run_data("fn"))
+
+    def test_get_filesystem_connection(self):
+        """
+        The default filesystem accessor.
+        """
+        inst = DSLibrary()
+        log = []
+        with mock.patch("dslibrary.front.connect_to_filesystem", lambda **k: log.append(k) or "FS"):
+            fs = inst.get_filesystem_connection("resource", for_write=True, arg1=123)
+            self.assertEqual(fs, "FS")
+        self.assertEqual(log, [{'uri': 'resource', 'for_write': True, 'arg1': 123}])
+
+    def test_get_sql_connection(self):
+        """
+        The default SQL accessor.
+        """
+        inst = DSLibrary()
+        log = []
+        with mock.patch("dslibrary.front.connect_to_database", lambda **k: log.append(k) or "NoSqlConn"):
+            fs = inst.get_sql_connection("resource", for_write=True, arg1=123)
+            self.assertEqual(fs, "NoSqlConn")
+        self.assertEqual(log, [{'uri': 'resource', 'library': None, 'for_write': True, 'arg1': 123}])
+
+    def test_get_nosql_connection(self):
+        """
+        The default NoSQL accessor.
+        """
+        inst = DSLibrary()
+        log = []
+        with mock.patch("dslibrary.front.connect_to_nosql", lambda **k: log.append(k) or "SqlConn"):
+            fs = inst.get_nosql_connection("resource", for_write=True, arg1=123)
+            self.assertEqual(fs, "SqlConn")
+        self.assertEqual(log, [{'uri': 'resource', 'library': None, 'for_write': True, 'arg1': 123}])
+
+    def test_log_mlflow(self):
+        inst = DSLibrary()
+        inst._mlflow_all = True
+        log = []
+        with mock.patch("mlflow.log_param", lambda k, v: log.append((k, v))):
+            inst.log_param("x", 1)
+        self.assertEqual(log, [("x", 1)])
+        log.clear()
+        with mock.patch("mlflow.log_metric", lambda *a: log.append(a)):
+            inst.log_metric("q", 222)
+            inst.log_metrics({"z": 9}, step=2)
+        self.assertEqual(log, [('q', 222, 0), ('z', 9, 2)])
+        log.clear()
+        with mock.patch("mlflow.log_dict", lambda *a: log.append(a)):
+            inst.log_dict({"x": 1}, "f.json")
+        self.assertEqual(log, [({'x': 1}, 'f.json')])
+        log.clear()
+        with mock.patch("mlflow.log_artifact", lambda *a: log.append(a)):
+            inst.log_artifact("f_local", "f_store")
+        self.assertEqual(log, [('f_local', 'f_store')])
+        log.clear()
+        # not a log function but close enough
+        self.assertRaises(DSLibraryException, lambda: inst.get_metrics("x"))
+
+    def test_mlflow_start_end(self):
+        inst = DSLibrary()
+        inst._mlflow_all = True
+        log = []
+        with mock.patch("mlflow.mlflow.start_run", lambda: log.append("start")):
+            with mock.patch("mlflow.mlflow.end_run", lambda: log.append("end")):
+                with mock.patch("mlflow.active_run", lambda: log.append("active_run")):
+                    # TODO mock this: mlflow.active_run().info.run_id
+                    with inst.start_run():
+                        pass
+        self.assertEqual(log, ['start', 'active_run', 'end'])
+
+    def test_yaml(self):
+        data = [
+            {"age": 41, "name": {"first": "John", "last": "Smith"}}
+        ]
+        class MyInst(DSLibrary):
+            def _opener(self, path: str, mode: str, **kwargs):
+                return io.StringIO(yaml.dump(data))
+        df = MyInst().load_dataframe("data.yaml")
+        assert list(df.age) == [41]
+        assert list(df["name.first"]) == ["John"]
+
+    def test_xlsx(self):
+        fn = os.path.join(os.path.dirname(__file__), "test_data/two_cols.xlsx")
+        class MyInst(DSLibrary):
+            def _opener(self, path: str, mode: str, **kwargs):
+                return open(path, mode)
+        df = MyInst().load_dataframe(fn)
+        self.assertEqual(list(df.a), [2, 7])
+        self.assertEqual(list(df.b), [5, 9])
+
+    def test_xlsx__unseekable_stream(self):
+        """
+        The xlsx parser requires a seekable stream, so when we supply an unseekable stream
+        it does some special processing.
+        """
+        fn = os.path.join(os.path.dirname(__file__), "test_data/two_cols.xlsx")
+        class MyInst(DSLibrary):
+            def _opener(self, path: str, mode: str, **kwargs):
+                with open(path, mode) as f_r:
+                    data = f_r.read()
+                class Stream(io.RawIOBase):
+                    def __init__(self, data):
+                        self.data = data
+                    def read(self, n=None):
+                        if n is None:
+                            n = len(self.data)
+                        out = self.data[:n]
+                        self.data = self.data[n:]
+                        return out
+                    def readable(self):
+                        return True
+                    def seekable(self):
+                        return False
+                    def __enter__(self):
+                        return self
+                    def __exit__(self, exc_type, exc_val, exc_tb):
+                        pass
+                return Stream(data)
+        df = MyInst().load_dataframe(fn)
+        self.assertEqual(list(df.a), [2, 7])
+        self.assertEqual(list(df.b), [5, 9])
