@@ -2,6 +2,7 @@ import unittest
 import pickle
 import tempfile
 import zipfile
+import io
 
 from dslibrary.transport.to_mmlibrary import DSLibraryViaMMLibrary
 
@@ -70,10 +71,11 @@ class TestToMMLibrary(unittest.TestCase):
 
     def test_get_sql_connection(self):
         class MyMM(object):
-            def get_db_connection(self, resource_name, database: str=None):
-                return f"{resource_name}:{database}"
+            def get_db_connection(self, resource_name):
+                return resource_name
         dsl = DSLibraryViaMMLibrary(_mm=MyMM())
-        assert dsl.get_sql_connection("x", database="d") == "x:d"
+        r = dsl.get_sql_connection("x", database="d")
+        self.assertEqual(r, "x")
 
     def test_load_pickled_model(self):
         with tempfile.NamedTemporaryFile() as f_tmp:
@@ -117,12 +119,15 @@ class TestToMMLibrary(unittest.TestCase):
         """
         The open_stream() method (available in DSW, not MM) can be used to access a file from a named filesystem.
         """
-        log = []
         class MyMM(object):
-            def open_stream(self, filename, file_source, mode):
-                log.append((filename, file_source, mode))
-                return "FH"
+            def get_db_connection(self, resource_name):
+                assert resource_name == "fs123"
+                class FH(object):
+                    def open(self, filename, mode, **kwargs):
+                        assert filename == "name"
+                        assert mode == "r"
+                        return io.StringIO("xyz")
+                return FH()
         dsl = DSLibraryViaMMLibrary(_mm=MyMM())
         r = dsl.open_resource("name", mode="r", filesystem="fs123", bucket="b1")
-        assert r == "FH"
-        assert log == [('name', 'fs123:b1', 'r')]
+        assert r.read() == "xyz"
